import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict

from src.chainsaw import featurisers
from src.chainsaw.domain_assignment.util import convert_domain_dict_strings
from src.chainsaw.factories import pairwise_predictor
from src.chainsaw.utils import common as common_utils
from src.shared.structure import get_model_structure

MODEL_DIR = Path(__file__).parent.parent.parent / 'model' / 'chainsaw'
LOG = logging.getLogger(__name__)


def load_model(
        remove_disordered_domain_threshold: float = 0.35,
        min_ss_components: int = 2,
        min_domain_length: int = 30,
        post_process_domains: bool = True, ):
    config = common_utils.load_json(os.path.join(MODEL_DIR, "config.json"))
    feature_config = common_utils.load_json(os.path.join(MODEL_DIR, "feature_config.json"))
    config["learner"]["remove_disordered_domain_threshold"] = remove_disordered_domain_threshold
    config["learner"]["post_process_domains"] = post_process_domains
    config["learner"]["min_ss_components"] = min_ss_components
    config["learner"]["min_domain_length"] = min_domain_length
    config["learner"]["dist_transform_type"] = config["data"].get("dist_transform", 'min_replace_inverse')
    config["learner"]["distance_denominator"] = config["data"].get("distance_denominator", None)
    learner = pairwise_predictor(config["learner"], output_dir=MODEL_DIR)
    learner.feature_config = feature_config
    learner.load_checkpoints()
    learner.eval()

    return learner


@dataclass
class PredictionResult:
    """
    Holds the result of a single domain prediction.
    """
    domain_id: str
    chopping: str
    confidence: float


class Seg:
    """
    Represents a contiguous segment of residues (by index).
    """

    def __init__(self, domain_id: str, start_index: int, end_index: int, index_to_label_map: Dict[int, str]):
        self.domain_id = domain_id
        self.start_index = start_index
        self.end_index = end_index
        self._index_to_label_map = index_to_label_map

    def _res_label_of_index(self, index: int) -> str:
        try:
            return self._index_to_label_map[index]
        except KeyError:
            raise ValueError(
                f"Index {index} not found in residue label map "
                f"({sorted(self._index_to_label_map.keys())})."
            )

    @property
    def start_label(self) -> str:
        return self._res_label_of_index(self.start_index)

    @property
    def end_label(self) -> str:
        return self._res_label_of_index(self.end_index)


class Dom:
    """
    Represents a collection of one or more Segs (which may be discontinuous).
    """

    def __init__(self, domain_id: str):
        self.domain_id = domain_id
        self.segs: List[Seg] = []

    def add_seg(self, seg: Seg) -> None:
        self.segs.append(seg)


# -----------------------------------------------------


def _predict(model, pdb_path, renumber_pdbs=True, pdbchain=None) -> str | None:
    """
    Makes the prediction and returns a list of PredictionResult objects
    """

    # get model structure metadata
    model_structure = get_model_structure(pdb_path)

    if pdbchain is None:
        # get all the chain ids from the model structure
        all_chain_ids = [c.id for c in model_structure.get_chains()]
        # take the first chain id
        pdbchain = all_chain_ids[0]

    model_residues = featurisers.get_model_structure_residues(model_structure, chain=pdbchain)
    model_res_label_by_index = {int(r.index): str(r.res_label) for r in model_residues}

    x = featurisers.inference_time_create_features(pdb_path,
                                                   feature_config=model.feature_config,
                                                   chain=pdbchain,
                                                   renumber_pdbs=renumber_pdbs,
                                                   model_structure=model_structure,
                                                   )

    _, domain_dict, _ = model.predict(x)
    # Convert 0-indexed to 1-indexed to match AlphaFold indexing:
    domain_dict = [{k: [r + 1 for r in v] for k, v in d.items()} for d in domain_dict]
    names_str, bounds_str = convert_domain_dict_strings(domain_dict[0])

    if names_str == "":
        names = bounds = ()
    else:
        names = names_str.split('|')
        bounds = bounds_str.split('|')

    assert len(names) == len(bounds)

    class Seg:
        def __init__(self, domain_id: str, start_index: int, end_index: int):
            self.domain_id = domain_id
            self.start_index = int(start_index)
            self.end_index = int(end_index)

        def res_label_of_index(self, index: int):
            if index not in model_res_label_by_index:
                raise ValueError(f"Index {index} not in model_res_label_by_index ({model_res_label_by_index})")
            return model_res_label_by_index[int(index)]

        @property
        def start_label(self):
            return self.res_label_of_index(self.start_index)

        @property
        def end_label(self):
            return self.res_label_of_index(self.end_index)

    class Dom:
        def __init__(self, domain_id, segs=None):
            self.domain_id = domain_id
            if segs is None:
                segs = []
            self.segs = segs

        def add_seg(self, seg: Seg):
            self.segs.append(seg)

    # gather choppings into segments in domains
    domains_by_domain_id = {}
    for domain_id, chopping_by_index in zip(names, bounds):
        if domain_id not in domains_by_domain_id:
            domains_by_domain_id[domain_id] = Dom(domain_id)
        start_index, end_index = chopping_by_index.split('-')
        seg = Seg(domain_id, start_index, end_index)
        domains_by_domain_id[domain_id].add_seg(seg)

    # sort domain choppings by the start residue in first segment
    domains = sorted(domains_by_domain_id.values(), key=lambda dom: dom.segs[0].start_index)

    # collect domain choppings as strings
    domain_choppings = []
    for dom in domains:
        # convert segments to strings
        segs_str = [f"{seg.start_label}-{seg.end_label}" for seg in dom.segs]
        segs_index_str = [f"{seg.start_index}-{seg.end_index}" for seg in dom.segs]
        LOG.info(f"Segments (index to label): {segs_index_str} -> {segs_str}")
        # join discontinuous segs with '_'
        domain_choppings.append('_'.join(segs_str))

    # join domains with ','
    chopping_str = ','.join(domain_choppings)

    num_domains = len(domain_choppings)
    if num_domains == 0:
        chopping_str = None

    return chopping_str


def predict_chopping_chainsaw_from_pdb(
        structure_file,
        post_process_domains=True,
        remove_disordered_domain_threshold=0.35,
        min_domain_length=30,
        min_ss_components=2,
        use_first_chain=True,
):
    pdb_chain_id = None if use_first_chain else 'A'

    model = load_model(
        remove_disordered_domain_threshold=remove_disordered_domain_threshold,
        min_ss_components=min_ss_components,
        min_domain_length=min_domain_length,
        post_process_domains=post_process_domains,
    )

    return _predict(model, structure_file, pdbchain=pdb_chain_id)



from pathlib import Path
from typing import Optional
from pydantic import BaseModel, validator

def chain_id_from_pdb_path(pdb_path: Path):
    return pdb_path.stem

class PredictionResult(BaseModel):

    pdb_path: Path
    confidence: float
    chain_id: Optional[str]
    sequence_md5: str
    ndom: int
    nres: int
    chopping: Optional[str]
    time_sec: Optional[float]

    @validator('chain_id', always=True, pre=True, allow_reuse=True)
    def set_chain_id(cls, v, values):
        if v is None:
            return chain_id_from_pdb_path(values['pdb_path'])
        return v



import os
import re

import numpy as np

from scipy.spatial import distance_matrix

import torch
import torch.nn.functional as F

from src.merizo.model.utils.frames import get_frames
from src.merizo.model.utils.pdb_parser import (
    select_from_mol,
    get_xyz,
    open_pdb,
    check_bb,
    check_alt_res
)


resndict = {'ALA': 'A', 'ARG': 'R', 'ASN': 'N', 'ASP': 'D', 'CYS': 'C',
            'GLN': 'Q', 'GLU': 'E', 'GLY': 'G', 'HIS': 'H', 'ILE': 'I',
            'LEU': 'L', 'LYS': 'K', 'MET': 'M', 'PHE': 'F', 'PRO': 'P',
            'SER': 'S', 'THR': 'T', 'TRP': 'W', 'TYR': 'Y', 'VAL': 'V',
            'PAD': 'X',
            # Histidine variants:
            'HID': 'H', 'HIP': 'H', 'HIE': 'H',
            # Cystein variants:
            'CYX': 'C',
            }

aa_trans = str.maketrans('ARNDCQEGHILKMFPSTWYVBJOUXZ-.',
                         'ABCDEFGHIJKLMNOPQRSTUUUUUUVV')

def generate_features_domain(file: str, device: torch.device, pdb_chain: str="A"):
    """Returns single, pair, rotation, translation and resi index features for a pdb.

    Args:
        file (str): path to the pdb target.
        device (torch.device): torch.device object.

    Returns:
        Tuple[
            torch.tensor,   s:  [1,n,20]        single repr features
            torch.tensor,   z:  [1,n,n,1]       pair repr features
            torch.tensor,   r:  [1,n,3,3]       Gram-Schmidt rotation matrices
            torch.tensor,   t:  [1,n,3]         Gram-Schmidt translation vectors
            torch.tensor,   ri: [1,n]           residue indices for ALiBi
            np.ndarray,     pdb                 numpy structured array containing pdb records
            np.ndarray,     b:  [n]             numpy array of b-factor values
        ]
    """

    pdb, ca, backbone = pdb_to_features(file, pdb_chain)

    seq = torch.tensor(encode_seq(ca["resn"], three_letter=True)).long()

    s = F.one_hot(seq, 20)
    z = torch.tensor(get_dmap(ca, missing=False)).unsqueeze(-1)

    ri = torch.tensor(ca["resi"])
    b = torch.tensor(ca['b'])

    backbone = select_from_mol([backbone], "resi", ca["resi"])
    r, t = get_frames(backbone)

    s = s.unsqueeze(0).float().to(device)
    z = z.unsqueeze(0).float().to(device)
    r = r.unsqueeze(0).float().to(device)
    t = t.unsqueeze(0).float().to(device)
    ri = ri.unsqueeze(0).float().to(device)

    return {'s': s, 'z': z, 'r': r, 't': t, 'ri': ri, 'pdb': pdb, 'b': b, 'nres': s.shape[1]}

def pdb_to_features(file, pdb_chain: str="A", resi=None):

    if os.path.exists(file):
        pdb = open_pdb(file, pdb_chain)
        pdb = check_bb(check_alt_res(pdb), missing=False)
        pdb = [np.sort(pdb[0], order='resi')]

        if resi is not None:
            pdb = select_from_mol(pdb, 'resi', resi)

        backbone = select_from_mol(pdb, 'n', ['N', 'CA', 'C', 'O'])[0]
        ca = select_from_mol(pdb, 'n', ['CA'])[0]

        return pdb[0], ca, backbone
    else:
        print("Cannot find model {}".format(os.path.basename(file)))

def pdb_to_fasta(pdb: np.ndarray) -> str:
    """_summary_

    Args:
        pdb (np.ndarray): _description_

    Returns:
        str: _description_
    """

    return ''.join([resndict[x] for x in pdb[pdb['n'] == 'CA']['resn']])

def cath_dom_str_to_resi(c):
    """ Retrieves the first and last residue from a residue range.
        Handles negative first or negative second ids """

    c = re.sub(r'[A-Z]', '', c).replace('(', '').replace(')', '')

    # If first number is negative, temp replace with +
    if c[0] == '-':
        c = list(c)
        c[0] = '+'
        c = ''.join(c)

    # If second number is negative, temp replace with +
    if '--' in c:
        c = c.replace('--', '-+')

    a, b = c.split("-")

    a = int(a.replace('+', '-'))
    b = int(b.replace('+', '-'))

    return a, b


def boundaries_to_res(boundaries, query=None, resi=None):
    lb_resi = []
    lb_id = []
    ranges = []
    residues = []

    for c in boundaries:
        a, b = cath_dom_str_to_resi(c)

        assert a < b, f"{query}: boundary {c} - residues {a} cannot be greater than {b}"

        if resi is not None:
            # Hack to offset residues if there is 1 missing from the start/end
            orig_a, counter = a, 1
            while a not in resi:
                counter += 1
                a += 1

                assert counter < 5, f"Cannot find residue a: {orig_a} for pdb {query} in: {resi}"

            orig_b, counter = b, 1
            while b not in resi:
                counter += 1
                b -= 1

                assert counter < 5, f"Cannot find residue b: {orig_b} for pdb {query} in: {resi}"

            lb_id.append((list(resi).index(a), list(resi).index(b)))

        lb_resi.append((a, b))
        ranges.extend([a, b])
        residues.extend(range(a, b+1))

    if resi is not None:
        min_max_id = (list(resi).index(min(ranges)),
                      list(resi).index(max(ranges)))
    else:
        min_max_id = None

    domains = {
        'domain_resi': lb_resi,
        'domain_id': lb_id,
        'min_max_resi': (min(ranges), max(ranges)),
        'min_max_id': min_max_id,
        'residues': residues,
    }

    return domains


def encode_seq(seq, three_letter=False):
    """ Encodes a single-letter protein sequence into integers """
    if three_letter:
        seq = [resndict[s] for s in seq]

    return (np.frombuffer(''.join(seq).translate(aa_trans).encode(
        'latin-1'), dtype=np.uint8) - ord('A'))  # .view(len(seq))


def get_dmap(mol, missing=True):
    """ Returns (distance matrix, resi list) for a molecule

    Inputs:
        mol: pdb in a numpy structured array
        missing: if True, returns distance map accounting for missing residues
                 if False, returns distance map without gaps
    """

    if missing:
        resi, resn, coords, missing_resi = get_xyz(mol, gaps=True)
        dm = distance_matrix(coords.T, coords.T)
        return resi, resn, coords, dm, missing_resi
    else:
        coords = get_xyz(mol, gaps=False)
        dm = distance_matrix(coords.T, coords.T)
        return dm


def get_hmap(mol):
    """ Return the backbone hydrogen bonding map """

    mol = mol[np.argsort(mol, order=['resi'])]
    n_atoms = mol[mol['n'] == 'N']
    o_atoms = mol[mol['n'] == 'O']

    ri = abs(n_atoms['resi'][..., None] - n_atoms['resi'][..., None, :])
    hmap_no = distance_matrix(
        get_xyz(n_atoms, gaps=False).T,
        get_xyz(o_atoms, gaps=False).T
    )

    hmap_no[(hmap_no >= 3.5) | (ri < 3)] = 0
    hmap_no[hmap_no != 0] = 1

    hmap_on = distance_matrix(
        get_xyz(o_atoms, gaps=False).T,
        get_xyz(n_atoms, gaps=False).T
    )

    hmap_on[(hmap_on >= 3.5) | (ri < 3)] = 0
    hmap_on[hmap_on != 0] = 1

    hmap = torch.cat((torch.tensor(hmap_no).unsqueeze(-1),
                      torch.tensor(hmap_on).unsqueeze(-1)), dim=-1)

    assert hmap.shape[0] == hmap.shape[1]

    return hmap


def cent_to_dist(
    cent, min_bin=0, max_bin=90, no_bins=10, inf=1e8
):
    lower = torch.linspace(min_bin, max_bin, no_bins)
    upper = torch.cat([lower[1:], lower.new_tensor([inf])], dim=-1)

    dg = cent.unsqueeze(-1).expand(-1, -1, no_bins)

    return ((dg > lower) * (dg < upper)).type(cent.dtype)

def dm_to_distogram(dm, min_bin=3.25, max_bin=50.75, no_bins=39, inf=1e8):
    lower = torch.linspace(min_bin, max_bin, no_bins)
    upper = torch.cat([lower[1:], lower.new_tensor([inf])], dim=-1)

    dg = dm.unsqueeze(-1).expand(-1, -1, -1, no_bins)
    return ((dg > lower) * (dg < upper)).float()



from torch import nn

from .ipa.ipa_encoder import ipa_block
from .ipa.nndef_ipa_primitives import Rigid, Rotation

from .posenc.alibi import AlibiPositionalBias
from .decoders.mask_decoder import MaskTransformer


class Merizo(nn.Module):
    def __init__(self):
        super(Merizo, self).__init__()

        self.no_classes = 20

        self.linear_s_in = nn.Linear(20, 512, bias=False)
        self.linear_z_in = nn.Linear(1, 32, bias=False)

        self.ipa = ipa_block()
        self.alibi = AlibiPositionalBias(heads=16, slope_factor=1)

        self.decoder_head = MaskTransformer(
            n_cls=self.no_classes,
            n_layers=10,
            n_heads=16,
            d_model=512,
            d_ff=512,
            drop_path_rate=0,
            dropout=0,
        )

    def forward(self, features, mask=None):

        s, z, r, t, ri = features['s'], features['z'], features['r'], features['t'], features['ri']

        if mask is not None:
            s = s[:, mask]
            z = z[:, mask][:, :, mask]
            r = r[:, mask]
            t = t[:, mask]
            ri = ri[:, mask]

        ipa_out, _ = self.ipa(
            s=self.linear_s_in(s),
            z=self.linear_z_in(z),
            rigids=Rigid(Rotation(r), t),
        )

        domain_ids, conf_res = self.decoder_head(
            ipa_out,
            bias=self.alibi(ri.squeeze(0), clip=True)
        )

        return domain_ids, conf_res



"""There are three configurable things: predictors/models, data, training/evaluation.

Only first two require factories.
"""
import os

import pandas as pd

import logging

from src.chainsaw.domain_chop import PairwiseDomainPredictor
from src.chainsaw.models.rosetta import trRosettaNetwork
from src.chainsaw.domain_assignment.assigners import SparseLowRank
from src.chainsaw.utils import common as common_utils


LOG = logging.getLogger(__name__)


def get_assigner(config):
    assigner_type = config["type"]
    if assigner_type == "sparse_lowrank":
        assigner = SparseLowRank(**config["kwargs"])
    else:
        return ValueError()
    return assigner


def get_model(config):
    model_type = config["type"]
    if model_type == "trrosetta":
        model = trRosettaNetwork(**config["kwargs"])
    else:
        return ValueError()
    return model


def pairwise_predictor(
    learner_config,
    force_cpu=False,
    output_dir=None,
    use_mps_if_available=False,
):
    model = get_model(learner_config["model"])
    assigner = get_assigner(learner_config["assignment"])
    device = common_utils.get_torch_device(
        force_cpu=force_cpu,
        use_mps_if_available=use_mps_if_available,
    )
    model.to(device)
    kwargs = {k: v for k, v in learner_config.items() if k not in ["model",
                                                                   "assignment",
                                                                   "save_every_epoch",
                                                                   "uncertainty_model"
                                                                   ]}
    LOG.info(f"Learner kwargs: {kwargs}")
    return PairwiseDomainPredictor(model, assigner, device, checkpoint_dir=output_dir, **kwargs)


def get_test_ids(label_path, feature_path, csv_path=None):
    ids = [id.split('.')[0] for id in set(os.listdir(label_path)).intersection(set(os.listdir(feature_path)))]
    if csv_path is not None:
        df = pd.read_csv(csv_path)
        ids = sorted(list(set(ids).intersection(set(df.chain_id))))
    return ids


def filter_plddt(df_path, ids, threshold=90):
    df = pd.read_csv(df_path)
    df = df[df.plddt > threshold]
    ids = [i for i in ids if i in df.casp_id.values]
    return ids



from torch import nn

from .nndef_ipa import InvariantPointAttention, StructureModuleTransition
from .nndef_ipa_primitives import LayerNorm, Linear

class ipa_block(nn.Module):
    def __init__(self):
        super().__init__()

        # ---- IPA config ----
        self.c_s = 512
        self.c_z = 32
        self.c_ipa = 512
        self.no_qk_points = 4
        self.no_v_points = 8

        self.no_blocks = 6
        self.no_heads = 16
        self.no_transition_layers = 1
        self.dropout_rate = 0.0

        self.layer_norm_z = LayerNorm(self.c_z)
        self.layer_norm_s = LayerNorm(self.c_s)
        self.linear_in = Linear(self.c_s, self.c_s)

        self.ipa = InvariantPointAttention(
            c_s=self.c_s,
            c_z=self.c_z,
            c_hidden=self.c_ipa,
            no_heads=self.no_heads,
            no_qk_points=self.no_qk_points,
            no_v_points=self.no_v_points,
        )

        self.ipa_dropout = nn.Dropout(self.dropout_rate)
        self.layer_norm_ipa = LayerNorm(self.c_s)

        self.transition = StructureModuleTransition(
            self.c_s,
            self.no_transition_layers,
            self.dropout_rate
        )

    def forward(
        self, s, z, rigids, mask=None):

        s = self.layer_norm_s(s)
        s = self.linear_in(s)
        
        z = self.layer_norm_z(z)
            
        mask = s.new_ones(s.shape[:-1]) if mask is None else mask

        for _ in range(self.no_blocks):
            s_, attn = self.ipa(s=s, z=z, r=rigids, mask=mask)
            
            s = s + s_  # [*, N, C_s]
            s = self.ipa_dropout(s)
            s = self.layer_norm_ipa(s)
            s = self.transition(s)

        return s, attn


import os

import Bio.PDB
import Bio.PDB.Structure


def get_model_structure(structure_path) -> Bio.PDB.Structure:
    """
    Returns the Bio.PDB.Structure object for a given PDB or MMCIF file
    """
    structure_id = os.path.split(structure_path)[-1].split('.')[0]
    if structure_path.endswith('.pdb'):
        structure = Bio.PDB.PDBParser().get_structure(structure_id, structure_path)
    elif structure_path.endswith('.cif'):
        structure = Bio.PDB.MMCIFParser().get_structure(structure_id, structure_path)
    else:
        raise ValueError(f'Unrecognized file extension: {structure_path}')
    model = structure[0]
    return model



import os
from pathlib import Path

REPO_ROOT = Path(__file__).parent
STRIDE_EXE = os.environ.get('STRIDE_EXE', str(REPO_ROOT / "stride" / "stride"))
PYMOL_EXE = "/Applications/PyMOL.app/Contents/MacOS/PyMOL"  # only required if you want to generate 3D images

_3to1 = {
    'ALA': 'A',
    'CYS': 'C',
    'ASP': 'D',
    'GLU': 'E',
    'PHE': 'F',
    'GLY': 'G',
    'HIS': 'H',
    'ILE': 'I',
    'LYS': 'K',
    'LEU': 'L',
    'MET': 'M',
    'ASN': 'N',
    'PRO': 'P',
    'GLN': 'Q',
    'ARG': 'R',
    'SER': 'S',
    'THR': 'T',
    'VAL': 'V',
    'TRP': 'W',
    'TYR': 'Y',
}




RES_BUILD_INFO = {
    'ALA': {
        'torsion-names': [],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB'],
        'single-letter': 'A'
    },
    'ARG': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-CD', 'CB-CG-CD-NE', 'CG-CD-NE-CZ'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'CD', 'NE', 'CZ'],
        'single-letter': 'R'
    },
    'ASN': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-OD1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'OD1', 'ND2'],
        'single-letter': 'N'
    },
    'ASP': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-OD1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'OD1', 'OD2'],
        'single-letter': 'D'
    },
    'CYS': {
        'torsion-names': ['N-CA-CB-SG'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'SG'],
        'single-letter': 'C'
    },
    'GLN': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-CD', 'CB-CG-CD-OE1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'CD', 'OE1', 'NE2'],
        'single-letter': 'Q'
    },
    'GLU': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-CD', 'CB-CG-CD-OE1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'CD', 'OE1', 'OE2'],
        'single-letter': 'E'
    },
    'GLY': {
        'torsion-names': [],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB'],
        'single-letter': 'G'
    },
    'HIS': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-ND1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'ND1', 'CD2', 'CE1', 'NE2'],
        'single-letter': 'H'
    },
    'ILE': {
        'torsion-names': ['N-CA-CB-CG1', 'CA-CB-CG1-CD1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG1', 'CG2', 'CD1'],
        'single-letter': 'I'
    },
    'LEU': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-CD1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'CD1', 'CD2'],
        'single-letter': 'L'
    },
    'LYS': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-CD', 'CB-CG-CD-CE', 'CG-CD-CE-NZ'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'CD', 'CE', 'NZ'],
        'single-letter': 'K'
    },
    'MET': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-SD', 'CB-CG-SD-CE'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'SD', 'CE'],
        'single-letter': 'M'
    },
    'PHE': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-CD1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'CD1', 'CD2', 'CE1', 'CE2', 'CZ'],
        'single-letter': 'F'
    },
    'PRO': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-CD'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'CD'],
        'single-letter': 'P'
    },
    'SER': {
        'torsion-names': ['N-CA-CB-OG'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'OG'],
        'single-letter': 'S'
    },
    'THR': {
        'torsion-names': ['N-CA-CB-OG1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'OG1', 'CG2'],
        'single-letter': 'T'
    },
    'TRP': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-CD1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'CD1', 'CD2', 'NE1', 'CE2', 'CE3', 'CZ2', 'CZ3', 'CH2'],
        'single-letter': 'W'
    },
    'TYR': {
        'torsion-names': ['N-CA-CB-CG', 'CA-CB-CG-CD1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG', 'CD1', 'CD2', 'CE1', 'CE2', 'CZ', 'OH'],
        'single-letter': 'Y'
    },
    'VAL': {
        'torsion-names': ['N-CA-CB-CG1'],
        'atom-names': ['N', 'CA', 'C', 'O', 'CB', 'CG1', 'CG2'],
        'single-letter': 'V'
    },
}

SPECIAL_AA = {
    'MSE': {
        'convert-to': 'MET',
        'atom-mapping': {'SE': 'SD'},
        'remove': [],
        'name': 'selenomethionine'
    },
    'SEC': {
        'convert-to': 'CYS',
        'atom-mapping': {'SE': 'SG'},
        'remove': [],
        'name': 'selenocysteine'
    },
    'CSD': {
        'convert-to': 'CYS',
        'atom-mapping': {},
        'remove': ['OD1', 'OD2'],
        'name': 'sulphinoalanine'
    },
    'PCA': {
        'convert-to': 'GLU',
        'atom-mapping': {'OE': 'OE1'},
        'remove': [],
        'name': 'pyroglutamic_acid'
    },
    'PYL': {
        'convert-to': 'LYS',
        'atom-mapping': {},
        'remove': ['N2', 'CA2', 'C2', 'O2', 'CB2', 'CG2', 'CD2', 'CE2'],
        'name': 'pyrrolysine'
    },
}

EXCLUDE_AA = ['ASX', 'GLX', 'UNK']



import numpy as np
from collections import Counter
from .build_info import SPECIAL_AA, EXCLUDE_AA


def open_pdb(file: str, pdb_chain: str="A", ignore_H: bool = False):
    """ Return the coordinates of a PDB file in a numpy structured array.
        Residues declared in 'exclude' will be ignored. Declared in 'specials'
        will be read even if the residue is a HETATM.
        Only the first model is read (up to the first ENDMDL record).

        Inputs:
            file        str         path to pdb file

        Outputs:
            molecules   np.ndarray  structured array containing 12 fields.
                                    see dtype declared below.
    """

    exclude = EXCLUDE_AA
    specials = list(SPECIAL_AA.keys())

    only_atoms = ['C', 'N', 'CA', 'O']

    dtype = [
        ('type', 'U6'), ('i', 'i4'), ('n', 'U4'), ('alt', 'U1'),
        ('resn', 'U3'), ('chain', 'U2'), ('resi', 'i4'), ('x', 'f8'),
        ('y', 'f8'), ('z', 'f8'), ('occ', 'f8'), ('b', 'f8'), ('conf', 'f8')
    ]

    molecules = []
    with open(file, 'r') as enspdbfile:
        mol = []
        pdbfile = enspdbfile.readlines()
        
        for i, line in enumerate(pdbfile):
            if line[:4] == 'ATOM' or (line[:6] == 'HETATM' and line[17:20].strip() in specials):
                if not np.char.isnumeric(line[12]) and line[17:20].strip() not in exclude:
                    n = line[12:16].strip()
                    resn = line[17:20].strip()

                    if resn in specials:
                        resn = SPECIAL_AA[resn]['convert-to']

                    if ignore_H and n[0] == 'H':
                        continue

                    if n in only_atoms:
                        if line[20:22].strip() == pdb_chain:
                            mol.append(
                                (
                                    line[:6].strip(),       # type, e.g. ATOM
                                    line[6:11],             # i, atom number
                                    n,                      # n, atom name
                                    line[16:17].strip(),    # alt, alternative id
                                    resn,                   # resn, residue name
                                    line[20:22].strip(),    # chain
                                    line[22:26],            # resi, residue id
                                    line[30:38],            # x coordinate
                                    line[38:46],            # y coordinate
                                    line[46:54],            # z coordinate
                                    line[54:60],            # occupancy
                                    line[60:66],            # b-factor
                                    0.                      # confidence 
                                )
                            )

            if line[:6] == 'ENDMDL' or line[:3] == 'END' or i == len(pdbfile)-1: # or line[:3] == 'TER':
                mol = np.array(mol, dtype=dtype)
                molecules.append(mol)
                mol = []
                break
    if len(molecules[0]) == 0:
        print("Chain ID given not present in PDB file")
        exit(128)   
    return molecules


def check_special_aa(res: np.ndarray):
    """ Simply maps some non-standard residues to standard ones e.g. MSE -> MET

        Inputs:
            res         np.ndarray      residue structured array.

        Outputs:
            res         np.ndarray      If residue is not listed in mapping,
                                        returns the original res array.
                                        Otherwise converts 'resn' and 'n' of
                                        atom arrays in the res object.
    """

    resn = res['resn'][0]
    if resn in SPECIAL_AA.keys():
        res['type'] = 'ATOM'
        res['resn'] = SPECIAL_AA[resn]['convert-to']

        for n in SPECIAL_AA[resn]['atom-mapping'].keys():
            tmp_res = res[res['n'] == n]
            tmp_res['n'] = SPECIAL_AA[resn]['atom-mapping'][n]
            res[res['n'] == n] = tmp_res

        to_remove = SPECIAL_AA[resn]['remove']
        if len(to_remove) > 0:
            res = np.array([r for r in res if r['n'] not in to_remove])

    return res


def special_aa_mapping(mol: np.ndarray):
    """ Checks for residues listed in SPECIAL_AA. Maps special to standard
        residues.

        Inputs:
            mol     np.ndarray      Molecule-level array. Loops over any
                                    special residues designated in SPECIAL_AA.

        Outputs:
            mol     np.ndarray      Returns mol array with amended residues
                                    using check_special_aa().
                                    Specifically alters atom 'type', 'resn', 'n'
                                    Returns original mol if no specials found.
    """

    spc_res = mol[np.isin(mol['resn'], list(SPECIAL_AA.keys()))]

    if len(spc_res) != 0:
        spc_resi = list(set(spc_res['resi']))

        new_residues = []
        for id in spc_resi:
            res = check_special_aa(spc_res[spc_res['resi'] == id])
            new_residues.append(res)

        mol = np.concatenate(
            (mol[~np.isin(mol['resn'], list(SPECIAL_AA.keys()))],
             np.concatenate(new_residues))
        )

        sort_id = np.argsort(mol, order=['i'])
        mol = mol[sort_id]

    return mol


def select_from_mol(mols, name, selection_list):
    """ Extract rows of mol with 'name' in 'selection_list'
        e.g. ['ASN','ASP'] from 'resn'
        e.g. ['CA'] from 'n'
        e.g. [2, 3, 4, 5] from 'resi'
        e.g. ['A','B'] from 'chain' """

    molecule = []
    for m in mols:
        molecule.append(m[np.isin(m[name], selection_list)])

    return molecule


def get_xyz(mol, gaps=True):
    """ Neater way to return xyz coordinates from mol """

    if gaps:

        missing = []
        start_res = mol[0]['resi']
        end_res = mol[-1]['resi']
        expected_range = range(start_res, end_res+1)

        mol_padded = []
        for p in expected_range:
            res = mol[mol['resi'] == p]
            if len(res) == 0:
                mol_padded.append((p, 'PAD', np.NAN, np.NAN, np.NAN))
                missing.append(p)
            else:
                mol_padded.append(
                    (res['resi'], res['resn'][0], res['x'], res['y'], res['z']))

        mol = np.array(mol_padded, dtype=[('resi', 'i4'),
                                          ('resn', 'U3'),
                                          ('x', 'f8'),
                                          ('y', 'f8'),
                                          ('z', 'f8')])

        resi = np.array(mol['resi'])
        resn = np.array(mol['resn'])
        coords = np.array([mol['x'], mol['y'], mol['z']])
        missing_resi = missing

        return resi, resn, coords, missing_resi
    else:
        return np.array([mol['x'], mol['y'], mol['z']])


def check_bb(mol, missing=False):
    """ Checks each residue for N-CA-C atoms.
        Removes any residues with missing atoms
        If missing == True, then also return any residues
        with missing backbone atoms.
    """
    missing_res = []
    new_model = []

    for m in mol:
        complete_bb = []
        missing_bb = []

        resi = set(m['resi'])
        for r in resi:
            res = m[m['resi'] == r]
            if res[res['n'] == 'N']:
                if res[res['n'] == 'CA']:
                    if res[res['n'] == 'C']:
                        if res[res['n'] == 'O']:
                            complete_bb.append(res)
                        else:
                            missing_bb.append(res)
                    else:
                        missing_bb.append(res)
                else:
                    missing_bb.append(res)
            else:
                missing_bb.append(res)

        new_model.append(np.concatenate(complete_bb))

        if len(missing_bb) != 0:
            missing_res.append(np.concatenate(missing_bb))

    if missing:
        return new_model, missing_res
    else:
        return new_model


def check_alt_res(mol):
    """ Checks for alternative residue locations using
    occupancy. If multiple sets of atoms present for a
    residue, selects set with highest occupancy.
    """

    new_model = []

    for m in mol:
        single_occ_model = []

        resi = set(m['resi'])
        for r in resi:

            res = m[m['resi'] == r]
            counts = Counter(res['n'])

            if counts.most_common(1)[0][1] > 1:
                for ele in counts:
                    res_sort = np.sort(res[res['n'] == ele], order='occ')
                    single_occ_model.append([res_sort[-1]])
            else:
                single_occ_model.append(res)

        new_model.append(np.concatenate(single_occ_model))

    return new_model



import logging
import os
from typing import List

import Bio.PDB
import numpy as np
import torch
from scipy.spatial import distance_matrix

from src.chainsaw import constants
from src.chainsaw.constants import _3to1
from src.chainsaw.utils.secondary_structure import renum_pdb_file, \
    calculate_ss, make_ss_matrix

LOG = logging.getLogger(__name__)




class Residue:
    def __init__(self, index: int, res_label: str, aa: str):
        self.index = int(index)
        self.res_label = str(res_label)
        self.aa = str(aa)

def get_model_structure_residues(structure_model: Bio.PDB.Structure, chain='A') -> List[Residue]:
    """
    Returns a list of residues from a given PDB or MMCIF structure
    """
    residues = []
    res_index = 1
    for biores in structure_model[chain].child_list:
        res_num = biores.id[1]
        res_ins = biores.id[2]
        res_label = str(res_num)
        if res_ins != ' ':
            res_label += str(res_ins)

        aa3 = biores.get_resname()
        if aa3 not in _3to1:
            continue

        aa = _3to1[aa3]
        res = Residue(res_index, res_label, aa)
        residues.append(res)

        # increment the residue index after we have filtered out non-standard amino acids
        res_index += 1

    return residues


def inference_time_create_features(pdb_path, feature_config, chain="A", *,
                                   model_structure: Bio.PDB.Structure=None,
                                   renumber_pdbs=True, stride_path=constants.STRIDE_EXE,
                                   ):
    if not model_structure:
        model_structure = get_model_structure(pdb_path)

    dist_matrix = get_distance(model_structure, chain=chain)
    if renumber_pdbs:
        output_pdb_path = pdb_path.replace('.pdb', '_renum.pdb').replace('.cif', '_renum.cif')
        renum_pdb_file(pdb_path, output_pdb_path)
    else:
        output_pdb_path = pdb_path
    ss_filepath = pdb_path + '_ss.txt'
    calculate_ss(output_pdb_path, chain, stride_path, ssfile=ss_filepath)
    helix, strand = make_ss_matrix(ss_filepath, nres=dist_matrix.shape[-1])
    if feature_config['ss_bounds']:
        end_res_val = -1 if feature_config['negative_ss_end'] else 1
        helix_boundaries = make_boundary_matrix(helix, end_res_val=end_res_val)
        strand_boundaries = make_boundary_matrix(strand, end_res_val=end_res_val)
    if renumber_pdbs:
        os.remove(output_pdb_path)
    os.remove(ss_filepath)
    LOG.info(f"Distance matrix shape: {dist_matrix.shape}, SS matrix shape: {helix.shape}")
    if feature_config['ss_bounds']:
        if feature_config['same_channel_boundaries_and_ss']:
            helix_boundaries[helix == 1] = 1
            strand_boundaries[strand == 1] = 1
            stacked_features = np.stack((dist_matrix, helix_boundaries, strand_boundaries), axis=0)
        else:
            stacked_features = np.stack((dist_matrix, helix, strand, helix_boundaries, strand_boundaries), axis=0)
    else:
        stacked_features = np.stack((dist_matrix, helix, strand), axis=0)
    stacked_features = stacked_features[None] # add batch dimension
    return torch.Tensor(stacked_features)


def distance_matrix(x):
    """Compute the distance matrix.

    Returns the matrix of all pair-wise distances.

    Parameters
    ----------
    x : (M, K) array_like
        Matrix of M vectors in K dimensions.
    p : float, 1 <= p <= infinity
        Which Minkowski p-norm to use.
    Returns
    -------
    result : (M, M) ndarray
        Matrix containing the distance from every vector in `x` to every vector
        in `x`.
    """
    x1 = x[:, np.newaxis, :]  # Expand x to 3D for broadcasting, shape (M, 1, K)
    x2 = x[np.newaxis, :, :]  # Expand x to 3D for broadcasting, shape (1, M, K)

    distance_matrix = np.sum(np.abs(x2 - x1) ** 2, axis=-1) ** 0.5

    return distance_matrix.astype(np.float16)


def get_distance(structure_model: Bio.PDB.Structure, chain='A'):
    alpha_coords = np.array([residue['CA'].get_coord() for residue in \
                             structure_model[chain].get_residues() if Bio.PDB.is_aa(residue) and \
                             'CA' in residue and residue.get_resname() in _3to1], dtype=np.float16)
    x = distance_matrix(alpha_coords)
    return x


def make_boundary_matrix(ss, end_res_val=1):
    """
    makes a matrix where  the boundary residues
    of the sec struct component are 1
    """
    ss_lines = np.zeros_like(ss)
    diag = np.diag(ss)
    if max(diag) == 0:
        return ss_lines
    padded_diag = np.zeros(len(diag) + 2)
    padded_diag[1:-1] = diag
    diff_before = diag - padded_diag[:-2]
    diff_after = diag - padded_diag[2:]
    start_res = np.where(diff_before == 1)[0]
    end_res = np.where(diff_after == 1)[0]
    ss_lines[start_res, :] = 1
    ss_lines[:, start_res] = 1
    ss_lines[end_res, :] = end_res_val
    ss_lines[:, end_res] = end_res_val
    return ss_lines
